/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2007
   derniere modification le 04/03/2007 */

#ifndef SCESCENE_H__
#define SCESCENE_H__

#include <SCE/interface/SCENode.h>

/*
  Gestionnaire de scenes.
  
  Morceau principal du moteur, gestion des objets dans la scene:
  - octree/quadtree
  - BSP
  - frustum culling
  - contribution culling
  - modeles 3D --> SCEModel
  - cameras
    - mouvements avances
    - multiples
    - matrice de projection
  - lumieres
  - interactions:
    - objets/lumieres
    - camera/lumieres (occlusion lighting, cf occlusion queries)
  - ajouts d'objets
  - ajouts de lumieres
  - changement d'etats:
    - changement des arbres de partitionnement a utiliser --> BSP, quadtree, ...
    - effet global a la scene --> HDR, bloom, glow, ...
    - fonctionnalites divers...
  - LOD --> textures, modeles, terrain, ...
  - tri/ordre d'affichage:
    - textures
    - shaders
    - modeles
    - materiaux
  - sauvegarde d'un effet special definit par
    l'utilisateur et utilisation a la demande
  - screenshot --> DevIL pour l'enregistrement
  - video --> Open*L... a voir sur dvp.net (OpenCV .. ?)
  - rendu multi-passes --> HDR, bloom, post-effect, ...
  - IA de traitement des donnees (cf ordre d'affichage)
  - {occlusion culling
  - {occlusion queries
  - scene hierarchique ...
  - ...
*/


#if 0
typedef struct sce_scene_object SCE_SSceneObject;
struct sce_scene_object
{
    void *obj;                  /* pointeur sur l'objet qui peut etre de
                                   n'importe quel type (lumiere, modele, ...) */
    
    unsigned int type;          /* type de l'objet (enumerateur) */
    
    int draw;                   /* booleen, determine si l'objet doit etre
                                   afficher lors du prochain rendu */
    
    /* bof... */
    #if 0
    int active;                 /* booleen, determine si l'objet est actif
                                   ou non, par exemple dans le cas d'une
                                   lumiere, placer active a zero eteindra
                                   la lumiere */
    #endif
    
    /* ces deux objets servirons pour des tests d'intersection entre les
       objet, et divers optimisation (frustum culling, octree, ...) */
    
    SCE_SBoundingBox *box;      /* 'cube' qui entoure l'objet (bounding box) */
    SCE_SBoundingSphere *sphere;/* 'sphere' qui englobe l'objet */
    
    unsigned int lod_maxlevel;  /* valeur maximale du niveau de detail */
    unsigned int lod_level;     /* niveau de detail courant */
    
    /* ... */
};
#endif

typedef struct sce_sscene SCE_SScene;
struct sce_sscene
{
    /* objets de la scene:
       - lumiere(s)
       - skybox
       - arbre(s) de partitionnement actif
       - camera(s)
       - modeles 3D
       - ...
     */
    SCE_SNode *rootnode;
};

/* initialise le gestionnaire de scenes */
int SCE_Init_Scene (void);
/* quitte le gestionnaire de scenes */
void SCE_Quit_Scene (void);

/* cree une nouvelle scene */
SCE_SScene* SCE_Scene_Create (void);
/* supprime une scene */
void SCE_Scene_Delete (SCE_SScene*);

#endif /* guard */
