/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 21/09/2007
   derniere modification : 20/10/2007 */

#ifndef __SCELIST_H__
#define __SCELIST_H__

#include <SCE/SCEMinimal.h>

typedef struct sce_slist SCE_SList;
typedef struct sce_slistiterator SCE_SListIterator;

struct sce_slistiterator
{
    SCE_SListIterator *next, *prev;
    void *data;
};

typedef void (*SCE_FListFreeFunc)(void*);
typedef int (*SCE_FListCompareData)(void*, void*);

struct sce_slist
{
    SCE_SListIterator *first;
    SCE_FListFreeFunc f;
};


/* initialise une liste */
void SCE_List_Init (SCE_SList*);


SCE_SList* SCE_List_Create (SCE_FListFreeFunc);
void SCE_List_Clear (SCE_SList*);
void SCE_List_Delete (SCE_SList*);


/* ajout d'un element... */
/* ...par localisation */
int SCE_List_AddFront (SCE_SList*, void*);
int SCE_List_AddBack (SCE_SList*, void*);
#if 0
/* ...par indice */
int SCE_List_Add (SCE_SList*, SCE_SListIterator*, void*);
#endif


/* suppression d'un element... */
/* ...par localisation */
void* SCE_List_RemoveFront (SCE_SList*);
void* SCE_List_RemoveBack (SCE_SList*);
/* ...par indice */
void* SCE_List_Remove (SCE_SList*, SCE_SListIterator*);

/* suppression d'un element _ET_ suppression des donnees */
/* ...par localisation */
void SCE_List_EraseFront (SCE_SList*);
void SCE_List_EraseBack (SCE_SList*);
/* ...par indice */
void SCE_List_Erase (SCE_SList*, SCE_SListIterator*);


/* retourne le nombre d'iterateur dans la liste */
unsigned int SCE_List_GetSize (SCE_SList*);
/* SCE_List_GetFirst (); */
/* renvoie le dernier iterateur d'une liste */
SCE_SListIterator* SCE_List_GetLast (SCE_SList*);
/* renvoie les donnees d'un iterateur */
void* SCE_List_GetData (SCE_SListIterator*);
/* renvoie l'index (position) d'un iterateur dans sa liste */
unsigned int SCE_List_GetIndex (SCE_SListIterator*);
/* renvoie l'iterateur indique par l'index, ou NULL s'il n'existe pas */
SCE_SListIterator* SCE_List_GetIterator (SCE_SList*, unsigned int);
/* recherche l'iterateur possedant les donnees specifiees */
SCE_SListIterator* SCE_List_LocateIterator (SCE_SList*, void*,
                                            SCE_FListCompareData);
/* recherche l'index de l'iterateur possedant les donnees specifiees */
unsigned int SCE_List_LocateIndex (SCE_SList*, void*,
                                   SCE_FListCompareData);

/* assigne la fonction de liberation des donnees des iterateurs d'une liste */
void SCE_List_SetFreeFunc (SCE_SList*, SCE_FListFreeFunc);

#define SCE_List_ForEach(it, l)\
    for ((it) = (l)->first; (it); (it) = (it)->next)

#define SCE_List_ForEachNext(it)\
    for (; (it); (it) = (it)->next)

#define SCE_List_ForEachPrev(it)\
    for (; (it); (it) = (it)->prev)

#endif /* guard */
