/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 15 decembre 2006
   derniere modification le 26/10/2007 */

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECore.h>


/* initialise opengl */ /* NOTE: elle est pas inutile cette fonction ??? */
static int SCE_CInitGL (void)
{
    /* GLenum error;*/

    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();

    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();

    /* ne joue que sur l'unite de texture active, donc la 1 */
    glMatrixMode (GL_TEXTURE);
    glLoadIdentity ();

    glMatrixMode (GL_MODELVIEW);

    glClearColor (0.0, 0.0, 0.0, 0.0);
    /*glClearIndex (0.0);*/ /* ... */
    glClearDepth (1.0);
    glClearStencil (0);
    glClearAccum (0.0, 0.0, 0.0, 0.0);

    glShadeModel (GL_SMOOTH);
/* ...
    error = glGetError ();

    if (error != GL_NO_ERROR)
    {
        Logger_Log (error);
        Logger_LogMsg ((char*)gluErrorString (error));
        return SCE_ERROR;
    }
*/
    return SCE_OK;
}


/* revise le 08/10/2007 */
int SCE_CInit (SCEflags flags)
{
    GLenum error;

    SCE_btstart ();
    if (SCE_CInitGL () < 0)
    {
        SCE_btend ();
        return SCE_ERROR;
    }

    error = glewInit ();
    if (error != GLEW_OK)
    {
        Logger_Log ((signed)error);
        Logger_LogMsg ("cannot initialize GLEW : %s",
                       (char*)glewGetErrorString (error));
        SCE_btend ();
        return SCE_ERROR;
    }

    /* initialisation des modules du coeur */
    if (SCE_CSupportInit () < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    if (SCE_CImageInit () < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    if (SCE_CTextureInit () < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    if (SCE_CShaderInit (flags & SCE_CINIT_CG_SHADERS) < 0)
    {
        Logger_LogSrc ();
        SCE_CImageQuit ();
        SCE_btend ();
        return SCE_ERROR;
    }

    /* hop, on set les couleurs par defaut */
    SCE_CUseMaterial (NULL);

    SCE_btend ();
    return SCE_OK;
}

void SCE_CQuit (void)
{
    SCE_btstart ();
    SCE_CShaderQuit ();
    SCE_CTextureQuit ();
    SCE_CImageQuit ();
    SCE_btend ();
}


void SCE_CClear (const GLbitfield mask)
{
    glClear (mask);
}


void SCE_CSetState (SCEenum state, int active)
{
    if (active)
        glEnable (state);
    else
        glDisable (state);
}

void SCE_CSetState2 (SCEenum state, SCEenum state2, int active)
{
    if (active)
    {
        glEnable (state);
        glEnable (state2);
    }
    else
    {
        glDisable (state);
        glDisable (state2);
    }
}

void SCE_CSetState3 (SCEenum state, SCEenum state2, SCEenum state3, int active)
{
    if (active)
    {
        glEnable (state);
        glEnable (state2);
        glEnable (state3);
    }
    else
    {
        glDisable (state);
        glDisable (state2);
        glDisable (state3);
    }
}

void SCE_CSetState4 (SCEenum state, SCEenum state2,
                     SCEenum state3, SCEenum state4, int active)
{
    if (active)
    {
        glEnable (state);
        glEnable (state2);
        glEnable (state3);
        glEnable (state4);
    }
    else
    {
        glDisable (state);
        glDisable (state2);
        glDisable (state3);
        glDisable (state4);
    }
}

void SCE_CSetState5 (SCEenum state, SCEenum state2, SCEenum state3,
                     SCEenum state4, SCEenum state5, int active)
{
    if (active)
    {
        glEnable (state);
        glEnable (state2);
        glEnable (state3);
        glEnable (state4);
        glEnable (state5);
    }
    else
    {
        glDisable (state);
        glDisable (state2);
        glDisable (state3);
        glDisable (state4);
        glDisable (state5);
    }
}

void SCE_CSetState6 (SCEenum state, SCEenum state2, SCEenum state3,
                     SCEenum state4, SCEenum state5, SCEenum state6, int active)
{
    if (active)
    {
        glEnable (state);
        glEnable (state2);
        glEnable (state3);
        glEnable (state4);
        glEnable (state5);
        glEnable (state6);
    }
    else
    {
        glDisable (state);
        glDisable (state2);
        glDisable (state3);
        glDisable (state4);
        glDisable (state5);
        glDisable (state6);
    }
}
